import numpy as np
from typing import List
import logging
from ..core import Factor

class Cluster:
    def __init__(self, factors: List[Factor]):
        self.factors = factors
        self.scope = sorted(list(set.union(*[set(f.scope) for f in self.factors])))
        self.edges = []
        self.messages = dict()  # incoming messages, where each edge is the key
        self.potential = self._get_potential()

    def add_edge(self, edge):
        assert set(edge.scope).issubset(set(self.scope))
        self.edges.append(edge)
        self.messages[edge] = Factor.get_null()
    
    def send_message(self, edge):
        target_node = edge.get_other_node(self)
        variables_to_marginalize = \
            sorted(list(set(self.scope) - set(edge.scope)))
        marginalized_belief = self.get_belief() \
            .marginalize(variables_to_marginalize)
        msg = marginalized_belief / self.messages[edge]
        target_node.messages[edge] = msg

    def __repr__(self):
        return "ψ(" + ", ".join([f"{s}" for s in self.scope]) + ")"
    
    def _get_potential(self):
        product = self.factors[0]
        for f in self.factors[1:]:
            product = product * f
        return product
    
    def get_belief(self):
        belief = self.potential
        for msg in self.messages.values():
            belief = belief * msg
        return belief


class ClusterEdge:
    def __init__(self, nodes: List[Cluster], scope: List[Factor]):
        assert len(nodes) == 2
        self.nodes = nodes
        self.scope = scope
        for n in self.nodes:
            n.add_edge(self)
    
    def get_other_node(self, node):
        return (set(self.nodes) - set([node])).pop()
    
    def __repr__(self):
        n0 = self.nodes[0]
        n1 = self.nodes[1]
        return f"{n0} <-- {self.scope} --> {n1}"


class ClusterGraph:
    """A cluster graph is used for computing marginal probability distributions. 
    It is an undirected graph with edges between clusters. """
    def __init__(self, nodes: List[Cluster]):
        self.nodes = nodes
        self.variables = self._get_variables()
        self.factors = self._get_factors()

    def _get_variables(self):
        variables = set()
        for n in self.nodes:
            for f in n.factors:
                variables.update(f.scope)
        return variables
    
    def get_variable(self, name: str):
        return [v for v in self.variables if v.name==name][0]
    
    def _get_factors(self):
        factors = set()
        for n in self.nodes:
            factors.update(n.factors)
        return factors


    def propagate_beliefs_round_robin(self, num_times:int):
        for _ in range(num_times):
            for cluster in self.nodes:
                for edge in cluster.edges:
                    cluster.send_message(edge)

    def forward_backward_algorithm(self):
        """ This only runs on chains with no branches """
        def make_pass(nodes):
            previous_node = None
            for node in nodes[:-1]:
                edge = node.edges[0]
                if previous_node == edge.get_other_node(node):
                    edge = node.edges[1]
                previous_node = node
                node.send_message(edge)
        make_pass(self.nodes) # forward pass
        make_pass(self.nodes[::-1]) # backward pass




import logging
from ..core import Factor, CG_Node, CG
import cgm


class DagVariableElimination:
    def __init__(self, cg: CG):
        self.cg = cg

    def variable_elimination(self, nodes_to_eliminate: list[CG_Node]):
        """Implements the variable elimination algorithm over a CG.
        Specify a list of nodes to eliminate, in order of elimination
        """
        factors: set[cgm.CPD] = set(n.cpd for n in self.cg.nodes)
        for n in nodes_to_eliminate:
            logging.debug("Eliminating %s", n)
            factors = self._eliminate_node(n, factors)
            logging.debug(factors)
        return factors

    def _eliminate_node(self, node_to_eliminate: CG_Node, factors: set[cgm.CPD]):
        factors_to_combine = [f for f in factors if node_to_eliminate in f.scope]
        intermediate_factor = factors_to_combine[0]
        for f in factors_to_combine[1:]:
            intermediate_factor = intermediate_factor * f
        marginal_factor = intermediate_factor.marginalize([node_to_eliminate])
        reduced_factors = [f for f in factors if f not in factors_to_combine] + [marginal_factor]
        return reduced_factors

    def get_scope_of_new_factor_after_variable_elimination(self):
        """Returns a dictionary containing the new factor created after 
        eliminating a node, for every node. To reduce computation, we want to
        create new intermediate factors that have the smallest scope.
        """
        nodes = self.cg.nodes
        return {n_target : set.union(*[set(n.cpd.scope) for n in nodes if n_target in n.cpd.scope]) \
                - set([n_target]) for n_target in nodes}




import numpy as np
import cgm

def get_cg1():
    np.random.seed(30)
    # Define all nodes
    A = cgm.CG_Node('A', 3)
    B = cgm.CG_Node('B', 3)
    C = cgm.CG_Node('C', 3)
    D = cgm.CG_Node('D', 3)
    E = cgm.CG_Node('E', 3)
    F = cgm.CG_Node('F', 3)
    # Specify all parents of nodes
    cgm.CPD(B, [A])
    cgm.CPD(C, [A])
    cgm.CPD(D, [B, C])
    cgm.CPD(F, [C])
    cgm.CPD(E, [D])
    graph = cgm.CG([A, B, C, D, E, F])
    return graph

def get_cg2():
    np.random.seed(30)
    # Define all nodes
    season    = cgm.CG_Node('season', 4)
    rain      = cgm.CG_Node('rain', 2)
    sprinkler = cgm.CG_Node('sprinkler', 2)
    wet       = cgm.CG_Node('wet', 2)
    slippery  = cgm.CG_Node('slippery', 2)

    # Specify all parents of nodes
    cgm.CPD(season, [], np.array([.25, .25, .25, .25]))
    cgm.CPD(rain, [season])
    cgm.CPD(wet, [rain, sprinkler])
    cgm.CPD(slippery, [wet])
    graph = cgm.CG([season, rain, sprinkler, wet, slippery])
    return graph


"""
This module provides a class for performing forward sampling on a graphical 
model.
"""
import numpy as np
import cgm

SampleDict = dict[cgm.CG_Node, int]
PartialSampleDict = dict[cgm.CG_Node, int|None]

class ForwardSampler:
    """
    A class for performing forward sampling on a graphical model.

    Attributes:
        cg (cgm.CG): The graphical model to perform forward sampling on.
        nodes (list[cgm.CG_Node]): The list of nodes in the graphical model.
        samples (cgm.Factor[cgm.CG_Node]): The factor representing the samples.
        rng (np.random.Generator): The random number generator.

    Example Usage:
    ```
    cg = cgm.example_graphs.get_cg1()
    sampler = cgm.ForwardSampler(cg, seed=30)
    samples = sampler.get_n_samples(100)
    print(samples.values)
    ```
    """

    def __init__(self, cg: cgm.CG, seed: int = 30):
        """
        Initializes a ForwardSampler object.

        Args:
            cg (cgm.CG): The graphical model to perform forward sampling on.
            seed (int): The seed for the random number generator.
        """
        self.cg: cgm.CG = cg
        self.nodes: list[cgm.CG_Node] = sorted(cg.nodes)
        self.samples: cgm.Factor[cgm.CG_Node] = cgm.Factor[cgm.CG_Node].get_null()
        self.rng = np.random.default_rng(seed)

    def get_n_samples(self, num_samples: int):
        """
        Generates n samples from the graphical model.

        Args:
            num_samples (int): The number of samples to generate.

        Returns:
            cgm.Factor[cgm.CG_Node]: The factor representing the generated samples.
        """
        scope: list[cgm.CG_Node] = sorted(self.cg.nodes)
        num_dims = tuple(n.num_states for n in self.cg.nodes)
        self.samples = cgm.Factor[cgm.CG_Node](scope, np.zeros(num_dims, dtype=int))
        for _ in range(num_samples):
            sample, self.rng = self._forward_sample(self.rng)
            idx: tuple[int, ...] = tuple(sample[v] for v in self.samples.scope)
            self.samples.increment_at_index(idx, 1)
        return self.samples

    def get_sampled_marginal(self, nodes: set):
        """
        Computes the marginal count of the sampled nodes.

        Args:
            nodes (set): The set of nodes to compute the marginal count for.

        Returns:
            cgm.Factor[cgm.CG_Node]: The factor representing the computed marginal count.
        """
        return self.samples.marginalize(list(set(self.cg.nodes) - nodes))

    def _forward_sample(self, rng: np.random.Generator) -> tuple[SampleDict, np.random.Generator]:
        """
        Performs forward sampling on the graphical model.

        Args:
            rng (np.random.Generator): The random number generator.

        Returns:
            SampleDict: The generated sample.
        """
        current_sample: PartialSampleDict = {n: None for n in self.cg.nodes}
        for n in self.cg.nodes:
            current_sample, rng = self._sample_node(n, current_sample, rng)
        populated_sample: SampleDict = {k: v for k, v in current_sample.items() if v is not None}
        return populated_sample, rng

    def _sample_node(self,
                     node: cgm.CG_Node,
                     current_sample: PartialSampleDict,
                     rng: np.random.Generator) -> tuple[PartialSampleDict, np.random.Generator]:
        """
        Samples a node in the graphical model.

        Args:
            node (cgm.CG_Node): The node to sample.
            rng (np.random.Generator): The random number generator.

        Returns:
            np.random.Generator: The updated random number generator.
        """
        for p in node.parents:
            if current_sample[p] is None:
                current_sample, rng = self._sample_node(p, current_sample, rng)
        populated_sample = {k: v for k, v in current_sample.items() if v is not None}
        parent_states = {n: populated_sample[n] for n in node.parents}
        conditioned_node = node.cpd.condition(parent_states)
        sample, rng = conditioned_node.sample(1, rng)
        current_sample[node] = sample.item()
        return current_sample, rng



"""
Inside the main directory, run with:
`python -m pytest tests/core_tests.py`
"""
import pytest
import numpy as np
import numpy.testing as npt
import cgm

def test_generate_graph1():
    graph = cgm.example_graphs.get_cg1()
    assert len(graph.nodes) == 6

def test_factor():
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    phi1 = cgm.Factor[cgm.Variable]([a, b], 
                                    np.ones((num_states_a, num_states_b)))
    
    assert phi1.scope == [a, b]
    assert phi1.values.shape == (num_states_a, num_states_b)
    margin = phi1.marginalize([a]).values
    npt.assert_allclose(margin, num_states_a * np.ones((num_states_b,)))
    margin = phi1.marginalize([b]).values
    npt.assert_allclose(margin, num_states_b * np.ones((num_states_a,)))

def test_factor_unsorted_scope():
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    phi1 = cgm.Factor[cgm.Variable]([b, a],
                                    np.ones((num_states_b, num_states_a)))
    assert phi1.scope == [b, a]
    assert phi1.values.shape == (num_states_b, num_states_a)
    margin = phi1.marginalize([a]).values
    npt.assert_allclose(margin, num_states_a * np.ones((num_states_b,)))
    margin = phi1.marginalize([b]).values
    npt.assert_allclose(margin, num_states_b * np.ones((num_states_a,)))

def test_factor_incorrect_input_dims():
    """This test should raise an error."""
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    with pytest.raises(cgm.ScopeShapeMismatchError):
        cgm.Factor[cgm.Variable]([a, b],
                                 np.ones((num_states_b, num_states_a)))
    phi1 = cgm.Factor[cgm.Variable]([a, b],
                                 np.ones((num_states_a, num_states_b)))
    scope_shape = np.array([s.num_states for s in phi1.scope])
    npt.assert_equal(scope_shape, np.array([num_states_a, num_states_b]))

def test_check_input_with_non_unique_names():
    """Test _check_input with non-unique variable names to ensure it raises 
    NonUniqueVariableNamesError."""
    num_states = 2
    var_name = 'a_duplicate'
    a1 = cgm.Variable(var_name, num_states)
    a2 = cgm.Variable(var_name, num_states)
    with pytest.raises(cgm.NonUniqueVariableNamesError) as exc_info:
        cgm.Factor[cgm.Variable]([a1, a2])
        assert var_name in str(exc_info.value)

def test_CG_Node():
    a = cgm.CG_Node('a', 2)
    assert a.name == 'a'
    assert a.num_states == 2

def test_CPD_2nodes():
    num_states_a = 2
    num_states_b = 3
    a = cgm.CG_Node('a', num_states_a)
    b = cgm.CG_Node('b', num_states_b)
    phi1 = cgm.CPD(a, [b])
    assert phi1.child == a
    assert phi1.parents == [b]
    assert a.parents == {b}
    margin = phi1.marginalize([phi1.child]).values
    npt.assert_allclose(margin, np.ones_like(margin))
    assert phi1.scope == [a, b]
    assert phi1.values.shape == (num_states_a, num_states_b)

def test_CPD_3nodes():
    num_states_a = 2
    num_states_b = 3
    num_states_c = 4
    a = cgm.CG_Node('a', num_states_a)
    b = cgm.CG_Node('b', num_states_b)
    c = cgm.CG_Node('c', num_states_c)
    phi1 = cgm.CPD(a, [b, c])
    assert phi1.child == a
    assert phi1.parents == [b, c]
    assert a.parents == {b, c}
    margin = phi1.marginalize([phi1.child]).values
    npt.assert_allclose(margin, np.ones_like(margin))
    assert phi1.scope == [a, b, c]
    assert phi1.values.shape == (num_states_a, num_states_b, num_states_c)

def test_CPD_condition():
    num_states_a = 2
    num_states_b = 3
    num_states_c = 4
    a = cgm.CG_Node('a', num_states_a)
    b = cgm.CG_Node('b', num_states_b)
    c = cgm.CG_Node('c', num_states_c)
    values_a1 = np.array([[.3, .8, .2, .7],
                          [.7, .2, .8, .1],
                          [.1, .5, .3, .9]])
    values = np.array([values_a1, 1 - values_a1])
    phi1 = cgm.CPD(a, [b, c], values=values)
    npt.assert_allclose(phi1.values, values)
    selected_b_index = 1
    parent_states = {b: selected_b_index}
    phi1_cond = phi1.condition(parent_states)
    assert phi1_cond.child == a
    assert phi1_cond.parents == [c]
    assert phi1_cond.values.shape == (num_states_a, num_states_c)
    expected_values = np.array([values_a1[selected_b_index, :],
                                1 - values_a1[selected_b_index, :]])
    npt.assert_allclose(phi1_cond.values, expected_values)

def test_factor_condition():
    num_states_a = 2
    num_states_b = 3
    num_states_c = 4
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    c = cgm.Variable('c', num_states_c)
    phi1 = cgm.Factor[cgm.Variable]([a, b, c], 5 * np.ones((num_states_a, 
                                                            num_states_b, 
                                                            num_states_c)))
    selected_b_index = 1
    phi1_cond = phi1.condition({b: selected_b_index})
    assert phi1_cond.scope == [a, c]
    assert phi1_cond.values.shape == (num_states_a, num_states_c)
    npt.assert_allclose(phi1_cond.values, 5 * np.ones((num_states_a, 
                                                       num_states_c)))
    
def test_factor_multiplication():
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    phi1 = cgm.Factor[cgm.Variable]([a, b], np.ones((num_states_a, num_states_b)))
    phi2 = cgm.Factor[cgm.Variable]([b, a], np.ones((num_states_b, num_states_a)))

    result = phi1 * phi2

    assert result.scope == [a, b]
    assert result.values.shape == (num_states_a, num_states_b)

def test_factor_multiplication_with_different_scopes():
    num_states_a = 2
    num_states_b = 3
    num_states_c = 4
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    c = cgm.Variable('c', num_states_c)
    phi1 = cgm.Factor[cgm.Variable]([a, b], np.ones((num_states_a, num_states_b)))
    phi2 = cgm.Factor[cgm.Variable]([b, c], np.ones((num_states_b, num_states_c)))

    result = phi1 * phi2

    assert result.scope == [a, b, c]
    assert result.values.shape == (num_states_a, num_states_b, num_states_c)
    assert np.allclose(result.values, np.ones((num_states_a, num_states_b, num_states_c)))

def test_factor_multiplication_with_nonintersection_scope():
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    phi1 = cgm.Factor[cgm.Variable]([a], np.ones((num_states_a,)))
    phi2 = cgm.Factor[cgm.Variable]([b], np.ones((num_states_b,)))

    result = phi1 * phi2

    assert result.scope == [a, b]
    assert result.values.shape == (num_states_a, num_states_b)
    assert np.allclose(result.values, np.ones((num_states_a, num_states_b)))
    
def test_factor_division_by_scalar():
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    phi = cgm.Factor[cgm.Variable]([a, b], np.ones((num_states_a, num_states_b)))

    result = phi / 2

    assert result.scope == [a, b]
    assert result.values.shape == (num_states_a, num_states_b)
    assert np.allclose(result.values, np.divide(np.ones((num_states_a, num_states_b)), 2))

def test_factor_division_by_factor():
    num_states_a = 2
    num_states_b = 3
    num_states_c = 4
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    c = cgm.Variable('c', num_states_c)
    phi1 = cgm.Factor[cgm.Variable]([a, b, c], np.ones((num_states_a, num_states_b, num_states_c)))
    phi2 = cgm.Factor[cgm.Variable]([b, c], np.ones((num_states_b, num_states_c)))

    result = phi1 / phi2

    assert result.scope == [a, b, c]
    assert result.values.shape == (num_states_a, num_states_b, num_states_c)
    assert np.allclose(result.values, np.divide(np.ones((num_states_a, num_states_b, num_states_c)), np.ones((num_states_b, num_states_c))))

def test_factor_argmax():
    num_states_a = 2
    num_states_b = 3
    a = cgm.Variable('a', num_states_a)
    b = cgm.Variable('b', num_states_b)
    phi = cgm.Factor[cgm.Variable]([a, b], np.array([[1, 2, 3], [4, 5, 6]]))

    result = phi.argmax(a)

    assert result.scope == [b]
    assert result.values.shape == (num_states_b,)
    assert np.allclose(result.values, np.array([1, 1, 1]))

    result = phi.argmax(b)

    assert result.scope == [a]
    assert result.values.shape == (num_states_a,)
    assert np.allclose(result.values, np.array([2, 2]))


# -*- coding: utf-8 -*-
"""
The core module contains the basic building blocks of a Causal Graphical Model.
"""
from typing import List, Sequence, TypeVar, Generic
import functools
import collections
import numpy as np
V = TypeVar('V', bound='Variable')  # V can be Variable or Variable's subclass
D = TypeVar('D', bound='DAG_Node')


class Variable:
    """A variable has a name and can taken on a finite number of states. 
    """

    def __init__(self, name: str, num_states: int):
        self._name = name
        self._num_states = num_states

    @property
    def name(self) -> str:
        """Return the name of the variable."""
        return self._name

    @property
    def num_states(self) -> int:
        """Return the number of states of the variable."""
        return self._num_states

    def __repr__(self):
        return self.name

    def __lt__(self, other):
        return self.name < other.name


class DAG_Node(Variable, Generic[D]):
    """A DAG (Directed Acyclic Graph) node is a variable in a DAG. 
    A node can have multiple parents and multiple children, but no cycles can be
    created.
    """

    def __init__(self, name: str, num_states: int):
        super().__init__(name, num_states)
        self._parents: set[D] = set()

    @property
    def parents(self) -> set[D]:
        """Return a set of all the parents of this node."""
        return self._parents

    @parents.setter
    def parents(self, parents: set[D]):
        """Set the parents of this node."""
        self._parents = parents
        if hasattr(self, 'ancestors'):
            del self.ancestors  # clear the cached property

    @functools.cached_property
    def ancestors(self) -> set[D]:
        """Return a set of all the ancestors of this node"""
        parents_remaining = self.parents.copy()
        ancestors_acc = set()
        while len(parents_remaining) > 0:
            node = parents_remaining.pop()
            ancestors_acc.add(node)
            ancestors_acc.update(node.ancestors)
            parents_remaining = parents_remaining - ancestors_acc
        return ancestors_acc


class CG_Node(DAG_Node['CG_Node']):
    """A Causal Graph Node

    A CG_Node is a variable in a Bayesian Network. 
    A node is associated with a single conditional probability 
    distribution (CPD), which is a distribution over the variable given its 
    parents. If the node has no parents, this CPD is a distribution over all the
    states of the variable. 

    Example:
    >>> A = cgm.CG_Node('A', 2)
    >>> B = cgm.CG_Node('B', 2)
    >>> C = cgm.CG_Node('C', 2)
    >>> phi1 = cgm.CPD(A, [B])
    >>> phi2 = cgm.CPD(B, [C])
    >>> phi3 = cgm.CPD(C, [])


    """

    def __init__(self, name: str, num_states: int):
        super().__init__(name, num_states)
        # by default the cpd has no parents; the node is unconnected
        self.cpd = CPD(self)

    @property
    def cpd(self) -> 'CPD':
        """Return the conditional probability distribution for this node."""
        return self._cpd

    @cpd.setter
    def cpd(self, cpd: 'CPD'):
        """Set the conditional probability distribution for this node."""
        self._cpd = cpd
        self.parents = set(self.cpd.scope) - set([self])


class ScopeShapeMismatchError(Exception):
    """Exception raised when the shape of a factor's scope does not match 
    the shape of its stored values array."""
    def __init__(self, expected_shape, actual_shape):
        message = f"Expected shape {expected_shape}, but got {actual_shape}. " \
                   "The factor's scope must correspond to the shape of the " \
                   "stored values array."
        super().__init__(message)

class NonUniqueVariableNamesError(Exception):
    """Exception raised when the variables in a factor's scope do not have unique names."""
    def __init__(self, non_unique_names):
        message = "Variables in the scope have non-unique names: "\
                  f"{non_unique_names}. All variables in the scope must have "\
                   "unique names."
        super().__init__(message)

class Factor(Generic[V]):
    """A factor is a function that has a list of variables in its scope, and 
    maps every combination of variable values to a real number. In this 
    implementation the mapping is stored as a np.ndarray. For example, if this
    factor's scope is the variables {A, B, C}, and each of these is a binary 
    variable, then to access the value of the factor for [A=1, B=0, C=1], the 
    entry can be accessed at self.values[1, 0, 1]. If the ndarray isn't 
    specified, a random one will be created. 

    All variables in the scope must have unique names. 

    Factors ϕ1 and ϕ2 can be multiplied and divided by ϕ1 * ϕ2 and ϕ1 / ϕ2. 
    A factor can be marginalized over a subset of its scope. For example, to 
    marginalize out variables A and B, call ϕ.marginalize([A, B]).

    Example:
    >>> A = cgm.Variable('A', 2)
    >>> B = cgm.Variable('B', 2)
    >>> C = cgm.Variable('C', 2)
    >>> phi1 = cgm.Factor([A, B, C])
    >>> phi2 = cgm.Factor([B, C])
    >>> phi3 = cgm.Factor([B, C])
    >>> phi1 * phi2
    >>> phi1 / phi2
    >>> phi1.marginalize([A, B])

    """

    def __init__(self,
                 scope: Sequence[V],
                 values: np.ndarray | None = None):
        self.scope = scope
        if values is None:
            self._values = self._get_random_values()
        else:
            self._values = values
        self._check_input()

    @classmethod
    def get_null(cls):
        """Return a factor with no scope and a single value of 1.0."""
        return cls[V](scope=[], values=np.float64(1))

    @property
    def values(self) -> np.ndarray:
        """Return the values of the factor."""
        return self._values

    def set_values(self, new_values: np.ndarray):
        """Set the values of the factor to new_values."""
        # the dimension of the factor cannot be changed using this method
        assert self._values.shape == new_values.shape
        self._values = new_values
        self._check_input()

    def randomize_values(self):
        """Set the values of the factor to random numbers between 0 and 1."""
        self._values = self._get_random_values()
        self._check_input()

    def _check_input(self):
        # all variable names have to be unique
        var_counts = collections.Counter([el.name for el in self.scope])
        if any([v > 1 for v in var_counts.values()]):
            non_unique_names = [k for k, v in var_counts.items() if v > 1]
            raise NonUniqueVariableNamesError(non_unique_names)
        assert len({s.name for s in self.scope}) == len(self.scope)

        # shape of scope much match shape of factor
        scope_shape = np.array([s.num_states for s in self.scope])
        if not np.array_equal(self.values.shape, scope_shape):
            raise ScopeShapeMismatchError(expected_shape=scope_shape,
                                          actual_shape=self.values.shape)

    def __repr__(self):
        return "ϕ(" + ", ".join([f"{s}" for s in self.scope]) + ")"

    def _get_random_values(self):
        num_dimensions = tuple(s.num_states for s in self.scope)
        return np.random.uniform(size=num_dimensions)

    def __mul__(self, other: 'Factor'):
        """ 
        Factor product as defined in PGM Definition 4.2 (Koller 2009)
        """
        scope1 = self.scope
        scope2 = other.scope
        scope_2_but_not_1 = [sc for sc in scope2 if sc not in scope1]
        result_scope = list(scope1) + scope_2_but_not_1
        scope2_padded = list(scope2) + [sc for sc in scope1 if sc not in scope2]
        arr1 = np.expand_dims(self.values, axis=tuple(range(len(scope1), len(result_scope))))
        arr2 = np.expand_dims(other.values, axis=tuple(range(len(scope2), len(result_scope))))
        destination_mapping = {value: index for index, value in enumerate(result_scope)}
        arr2_dst = [destination_mapping[element] for element in scope2_padded]
        arr2 = np.moveaxis(arr2, source=range(len(result_scope)), destination=arr2_dst)
        return Factor(result_scope, np.multiply(arr1, arr2))

    def __truediv__(self, other: 'Factor'):
        scope1 = self.scope
        if isinstance(other, (int, float)):
            return Factor(scope1, np.divide(self.values, other))
        scope2 = other.scope
        # The scope of the denominator must be a subset of that of the numerator
        assert set(scope1).intersection(scope2) == set(scope2)
        result_scope = list(scope1)
        scope2_padded = list(scope2) + [sc for sc in scope1 if sc not in scope2]
        arr1 = self.values
        arr2 = np.expand_dims(other.values, axis=tuple(range(len(scope2), len(result_scope))))
        destination_mapping = {value: index for index, value in enumerate(result_scope)}
        arr2_dst = [destination_mapping[element] for element in scope2_padded]
        arr2 = np.moveaxis(arr2, source=range(len(result_scope)), destination=arr2_dst)
        return Factor(result_scope, np.divide(arr1, arr2))

    def marginalize(self, variables: List[Variable]):
        """ 
        Sum over all possible states of a list of variables
        example: phi3.marginalize([A, B]) 
        """
        axes = tuple(np.where([s in variables for s in self.scope])[0])
        reduced_scope = [s for s in self.scope if s not in variables]
        return Factor(reduced_scope, np.sum(self._values, axis=axes))
    
    def argmax(self, variable: V):
        """ 
        Find the state of the variables that maximizes the factor
        example: phi3.argmax([A, B]) 
        """
        axis = self.scope.index(variable)
        reduced_scope = [s for s in self.scope if s != variable]
        max_indices = np.argmax(self._values, axis=axis)
        return Factor[V](reduced_scope, max_indices)

    def normalize(self):
        """Returns a factor with the same distribution whose sum is 1"""
        return Factor(self.scope, (self / self.marginalize(self.scope)).values)

    def increment_at_index(self, index: tuple[int, ...], amount):
        """Increment the value of the factor at a particular index by amount."""
        self._values[index] += amount

    def condition(self, condition_dict: dict[V, int]) -> 'Factor':
        """Condition on a set of variables.

        Condition on a set of variables at particular values of those variables.
        condition_dict is a dictionary where each key is a variable to condition 
        on and the value is an integer representing the index to condition on. 

        The scope of the returned factor will exclude all the variables 
        conditioned on. 
        """

        specified_vars = set(condition_dict.keys())
        new_scope = [v for v in self.scope if v not in specified_vars]
        assert specified_vars.issubset(set(self.scope))
        index: list[int | slice] = []
        for var in self.scope:
            if var in specified_vars:
                index.append(condition_dict[var])
            else:
                index.append(slice(None))
        index_tuple = tuple(index)
        cond_values = self.values[index_tuple]
        return Factor(list(new_scope), cond_values)


class CPD(Factor[CG_Node]):
    """Conditional Probability Distribution

    This is a type of factor with additional constraints. One variable in its
    scope is the child node, the others are the parents. The CPD must sum to 1
    for every particular value of the child node. Additionally, the CPD cannot
    introduce cycles in the DAG.

    Example:
    >>> A = cgm.CG_Node('A', 2)
    >>> B = cgm.CG_Node('B', 2)
    >>> C = cgm.CG_Node('C', 2)
    >>> phi1 = cgm.CPD(A, [B])
    >>> phi2 = cgm.CPD(B, [C])
    >>> phi3 = cgm.CPD(C, [])

    """

    def __init__(self,
                 child: CG_Node,
                 parents: Sequence[CG_Node] | None = None,
                 values: np.ndarray | None = None):
        self._child = child
        if parents is None:
            parents = []
        self._parents = parents
        scope: list[CG_Node] = [child] + list(parents)
        super().__init__(scope, values)
        self.child = child
        self._normalize()

    @property
    def child(self) -> CG_Node:
        """Return the child node of the CPD."""
        return self._child

    @child.setter
    def child(self, child: CG_Node):
        self._child = child
        self._nocycles()
        child.cpd = self

    @property
    def parents(self) -> Sequence[CG_Node]:
        """Return the parents of the CPD."""
        return self._parents

    def _nocycles(self):
        child = self.child
        parents = set(self.scope) - set([child])
        if len(parents) == 0:
            return
        ancestors = set.union(*[p.ancestors for p in parents])
        assert child not in ancestors

    def _normalize(self):
        # Normalize so it is a distribution that sums to 1
        normalized_values = (self / self.marginalize([self.child])).values
        self._values = normalized_values
        margin = self.marginalize([self.child]).values
        np.testing.assert_allclose(margin, np.ones_like(margin))

    def normalize(self):
        msg = "CPD has no method 'normalize', since it is already normalized."
        raise AttributeError(msg)

    def randomize_values(self):
        super().randomize_values()
        self._normalize()

    def set_values(self, new_values: np.ndarray):
        super().set_values(new_values)
        self._normalize()

    def sample(self,
               num_samples: int,
               rng: np.random.Generator) -> tuple[np.ndarray, 
                                                  np.random.Generator]:
        """Sample from the distribution"""
        samples = rng.choice(a=len(self.values),
                             size=num_samples,
                             p=self.values)
        return samples, rng

    def condition(self, condition_dict: dict[CG_Node, int]) -> 'CPD':
        """Condition on a set of variables.

        Condition on a set of variables at particular values of those variables.
        condition_dict is a dictionary where each key is a variable to condition 
        on and the value is an integer representing the index to condition on. 

        The scope of the returned factor will exclude all the variables 
        conditioned on. 
        """

        specified_parents = set(condition_dict.keys())
        unspecified_parents = set(self.parents) - specified_parents
        assert specified_parents.issubset(set(self.parents))
        index: list[int | slice] = []
        for var in self.scope:
            if var in specified_parents:
                index.append(condition_dict[var])
            else:
                index.append(slice(None))
        index_tuple = tuple(index)
        cond_values = self.values[index_tuple]
        return CPD(child=self.child,
                   parents=list(unspecified_parents),
                   values=cond_values)


class DAG(Generic[D]):
    """Directed Acyclic Graph"""

    def __init__(self, nodes: Sequence[D]):
        self.nodes = sorted(nodes)

    def __repr__(self):
        s = ''
        for n in self.nodes:
            parents = sorted(list(n.parents))
            s += f"{n} <- {parents}\n"
        return s


class CG(DAG[CG_Node]):
    """Causal Graph
    Contains a list of CG_Nodes. The information about connectivity is stored 
    at each node.
    """

    def __init__(self, nodes: list[CG_Node]):
        super().__init__(nodes)



"""
Inside 'tests' director, run with:
`python -m unittest inference_tests`
"""
import logging
import numpy as np
import numpy.testing as npt
import cgm
logging.basicConfig(level=logging.INFO)

def test_elimination1():
    logging.debug("Running test_elimination1()")
    cg = cgm.example_graphs.get_cg1()
    A, B, C, D, E, F = cg.nodes
    elimination_order = [F, E, D, C, B, A]
    dve = cgm.inference.variable_elimination.DagVariableElimination(cg)
    factors = dve.variable_elimination(elimination_order)
    npt.assert_almost_equal(factors[0].values, 1, decimal=5)
    
    
def test_elimination2():
    logging.debug(f"Running test_elimination2()")
    cg = cgm.example_graphs.get_cg2()
    rain, season, slippery, sprinkler, wet = cg.nodes
    elimination_order = [season, rain, sprinkler, wet]
    dve = cgm.inference.variable_elimination.DagVariableElimination(cg)
    slippery_factor = dve.variable_elimination(elimination_order)[0] 
    logging.debug(f"Remaining factor: {slippery_factor}")
    logging.debug(f"Probabilities of slippery: {slippery_factor.values}")
    npt.assert_almost_equal(slippery_factor.values.sum(), 1, decimal=5)


def generate_clustergraph1():
    """ From Coursera PGM Course 2, Week 2, Video: Belief Propagation 
    Algorithm, Example Cluster Graph (12 minute mark)."""
    np.random.seed(30)
    A = cgm.CG_Node('A', 3)
    B = cgm.CG_Node('B', 3)
    C = cgm.CG_Node('C', 3)
    D = cgm.CG_Node('D', 3)
    E = cgm.CG_Node('E', 3)
    F = cgm.CG_Node('F', 3)
    phi1 = cgm.Factor([A, B, C])
    phi2 = cgm.Factor([B, C])
    phi3 = cgm.Factor([B, D])
    phi4 = cgm.Factor([D, E])
    phi5 = cgm.Factor([B, E])
    phi6 = cgm.Factor([B, D])
    phi7 = cgm.Factor([B, D, F])
    psi1 = cgm.inference.Cluster([phi1])
    psi2 = cgm.inference.Cluster([phi2, phi3, phi6])
    psi3 = cgm.inference.Cluster([phi7])
    psi4 = cgm.inference.Cluster([phi5])
    psi5 = cgm.inference.Cluster([phi4])
    cgm.inference.ClusterEdge([psi1, psi2], [C])
    cgm.inference.ClusterEdge([psi1, psi4], [B])
    cgm.inference.ClusterEdge([psi2, psi4], [B])
    cgm.inference.ClusterEdge([psi2, psi5], [D])
    cgm.inference.ClusterEdge([psi3, psi4], [B])
    cgm.inference.ClusterEdge([psi3, psi5], [D])
    cgm.inference.ClusterEdge([psi4, psi5], [E])
    return cgm.inference.ClusterGraph([psi1, psi2, psi3, psi4, psi5])

def generate_clustergraph_chain():
    """ From Coursera PGM Course 2, Week 2, Video: Clique Tree Algorithm 
    - Correctness, Message Passing In Trees (2 minute mark)."""
    np.random.seed(30)
    A = cgm.CG_Node('A', 3)
    B = cgm.CG_Node('B', 3)
    C = cgm.CG_Node('C', 3)
    D = cgm.CG_Node('D', 3)
    E = cgm.CG_Node('E', 3)
    phi1 = cgm.CPD(A, [B])
    phi2 = cgm.CPD(B, [C])
    phi3 = cgm.CPD(C, [D])
    phi4 = cgm.CPD(D, [E])
    phi5 = cgm.CPD(E, [])
    psi1 = cgm.inference.Cluster([phi1])
    psi2 = cgm.inference.Cluster([phi2])
    psi3 = cgm.inference.Cluster([phi3])
    psi4 = cgm.inference.Cluster([phi4])
    psi5 = cgm.inference.Cluster([phi5])
    cgm.inference.ClusterEdge([psi1, psi2], [B])
    cgm.inference.ClusterEdge([psi2, psi3], [C])
    cgm.inference.ClusterEdge([psi3, psi4], [D])
    cgm.inference.ClusterEdge([psi4, psi5], [E])
    g = cgm.inference.ClusterGraph([psi1, psi2, psi3, psi4, psi5])
    return g

def test_clustergraph1():
    logging.debug('Creating Variables and Clusters in clusterGraphTest()')
    g = generate_clustergraph1()
    g.propagate_beliefs_round_robin(3)

def test_forward_backward_on_chain():
    logging.debug('Creating Variables and Clusters in test_cliquetree()')
    g = generate_clustergraph_chain()
    psi1 = g.nodes[0]
    A = g.get_variable('A')
    B = g.get_variable('B')
    logging.debug(f'\n{psi1.get_belief().values}')
    g.forward_backward_algorithm()
    p_of_A_estimate = psi1.get_belief().marginalize([B]).normalize().values
    logging.debug(f"Estimated distribution over A: \n{p_of_A_estimate}")
    joint_factor = cgm.Factor.get_null()
    for f in g.factors:
        joint_factor = joint_factor * f
    marginal_factor = joint_factor.marginalize(g.variables - {A})
    p_of_A_true = marginal_factor.values
    logging.debug(f"True distribution over A: \n{p_of_A_true}")
    np.testing.assert_array_almost_equal(p_of_A_true, p_of_A_estimate, decimal=2)

def test_roundrobin_on_chain():
    logging.debug('Creating Variables and Clusters in test_cliquetree()')
    g = generate_clustergraph_chain()
    psi1 = g.nodes[0]
    A = g.get_variable('A')
    B = g.get_variable('B')
    logging.debug(f'\n{psi1.get_belief().values}')
    g.propagate_beliefs_round_robin(5)
    p_of_A_estimate = psi1.get_belief().marginalize([B]).normalize().values
    logging.debug(f"Estimated distribution over A: \n{p_of_A_estimate}")
    joint_factor = cgm.Factor.get_null()
    for f in g.factors:
        joint_factor = joint_factor * f
    marginal_factor = joint_factor.marginalize(g.variables - {A})
    p_of_A_true = marginal_factor.values
    logging.debug(f"True distribution over A: \n{p_of_A_true}")
    np.testing.assert_array_almost_equal(p_of_A_true, p_of_A_estimate, decimal=2)

def test_forward_sample():
    logging.debug('Testing forward_sample()')
    cg = cgm.example_graphs.get_cg2()
    rain, season, slippery, sprinkler, wet = cg.nodes
    sampler = cgm.inference.forward_sampling.ForwardSampler(cg, 30)
    num_samples = 1000
    samples = sampler.get_n_samples(num_samples)
    marginal = sampler.get_sampled_marginal({season})
    expected_season = np.array([.25, .25, .25, .25])
    npt.assert_allclose(marginal.normalize().values,
                        expected_season,
                        rtol=.2)
    logging.debug(f'Season count: {marginal.normalize().values}')
